# -*- coding: utf8 -*-
from __future__ import unicode_literals

import time

from driverng import Worklist
from driverng.astm import AstmMessageV2, HeaderSegment
from driverng.astm import HprimDate as HL7Date
from driverng.astm import HprimForeign as HL7Foreign
from driverng.astm import HprimString as HL7String
from driverng.astm import Segment


class Cobas6800SegmentMSH(HeaderSegment):
    sending_application = HL7String((3, 1), default="LIS")
    sending_facility = HL7String((4, 1), default="LIS Facility")
    receiving_application = HL7String((5, 1), default="COBAS6800/8800")
    receiving_facility = HL7String((6, 1), default="Lab")
    date = HL7Date(7)
    message_type = HL7String((9, 1), default="RSP")
    event_type = HL7String((9, 2), default="K11")
    control_id = HL7String(10)
    version = HL7String(12, default="2.5")
    character_set = HL7String(18, default="UNICODE UTF-8")
    message_profile_identifier = HL7String((21, 1), default="ROC-03")
    message_profile_identifier_type = HL7String((21, 2), default="ROCHE")

    specimen = HL7Foreign("SPM", "header")
    specimen_container = HL7Foreign("SAC", "header")
    queries = HL7Foreign("QPD", "header")
    orders = HL7Foreign("OBR", "header")
    equals = HL7Foreign("EQU", "header")


class Cobas6800SegmentSPM(Segment):
    sample_id = HL7String(3)
    sample_identifier = HL7String((5, 1))
    sample_text = HL7String((5, 2))
    sample_name_coding = HL7String((5, 3))
    sample_role = HL7String(12, default="P")

    orders = HL7Foreign("OBR", "patient")
    specimen_container = HL7Foreign("SAC", "patient")


class Cobas6800SegmentSAC(Segment):
    secondary_sample_id = HL7String(3)
    accession_identifier = HL7String(4)
    container_identifier = HL7String((9, 1))
    container_name = HL7String((9, 3), default="HL70370")
    volume_input = HL7String(21)
    volume_units_identifier = HL7String((24, 1))
    volume_units_name = HL7String((24, 3))

    orders = HL7Foreign("OBR", "patient")
    common_order = HL7Foreign("ORC", "patient")
    inventory = HL7Foreign("INV", "result")
    specimens = HL7Foreign("SPM", "patient")


class Cobas6800SegmentOBR(Segment):
    place_order_id = HL7String(3)
    identifier = HL7String((4, 1))
    test = HL7String((4, 2))
    name_test = HL7String((4, 3))
    sample_action_code = HL7String(12)

    infos = HL7Foreign("ORC", "orders")
    results = HL7Foreign("OBX", "orders")
    infos_tests = HL7Foreign("TCD", "order")


class Cobas6800SegmentTCD(Segment):
    identifier = HL7String((2, 1))
    test = HL7String((2, 2))
    test_name = HL7String((2, 3))
    pool_size_num1 = HL7String((3, 2))
    pool_size_separator = HL7String((3, 3))
    pool_size_num2 = HL7String((3, 4))


class Cobas6800SegmentORC(Segment):
    order_control = HL7String(3)
    order_status = HL7String(6)
    transaction_date = HL7Date(10)

    orders = HL7Foreign("OBR", "patient")


class Cobas6800SegmentQPD(Segment):
    query_identifier = HL7String((2, 1))
    query_text = HL7String((2, 2))
    query_name = HL7String((2, 3))
    query_tag = HL7String(3)
    sample_identification = HL7String(4)
    rack_id = HL7String(6)
    position = HL7String(7)


class Cobas6800SegmentQAK(Segment):
    query_response = HL7String(3)


class Cobas6800SegmentOBX(Segment):
    value_type = HL7String(3)
    observation_identifier = HL7String((4, 1))
    observation_test = HL7String((4, 2))
    observation_name = HL7String((4, 3))
    observation_sub_id = HL7String(5)
    value = HL7String(6)
    units_identifier = HL7String((7, 1))
    units_name = HL7String((7, 3))
    interpretation_code = HL7String(8)
    result_status = HL7String(12)
    responsible_observer = HL7String(17)
    equipment_name = HL7String((19, 1))
    manufacturer = HL7String((19, 2))
    analysis_date = HL7Date(20)
    control_identifier = HL7String((29, 1))
    control_name = HL7String((29, 3))

    inventory = HL7Foreign("INV", "result")
    infos = HL7Foreign("NTE", "result")


class Cobas6800SegmentNTE(Segment):
    comment = HL7String(4)


class Cobas6800SegmentEQU(Segment):
    equipment = HL7String((2, 1))
    namespace_id = HL7String((2, 2))
    event_date = HL7Date(3)
    equipment_state = HL7String(4)


class Cobas6800SegmentINV(Segment):
    identifier = HL7String(2)
    status = HL7String(3)
    ressource_type = HL7String(4)
    quantity = HL7String(9)
    quantity_units = HL7String((12, 1))
    quantity_units_name = HL7String((12, 3))
    fluid_identifier = HL7String((16, 1))
    fluid_text = HL7String((16, 2))
    fluid_name = HL7String((16, 3))
    equipment_state_code = HL7String(22)
    equipment_state_value = HL7String((23, 1))
    equipment_state_capacity = HL7String((23, 2))


class Driver(AstmMessageV2):
    SPEC = {"code": "cobas6800", "label": "Cobas6800 HL7"}

    SEGMENT_CLASS_MAP = {
        "MSH": Cobas6800SegmentMSH,
        "SPM": Cobas6800SegmentSPM,
        "SAC": Cobas6800SegmentSAC,
        "ORC": Cobas6800SegmentORC,
        "OBR": Cobas6800SegmentOBR,
        "OBX": Cobas6800SegmentOBX,
        "TCD": Cobas6800SegmentTCD,
        "QPD": Cobas6800SegmentQPD,
        "QAK": Cobas6800SegmentQAK,
        "NTE": Cobas6800SegmentNTE,
        "EQU": Cobas6800SegmentEQU,
        "INV": Cobas6800SegmentINV,
    }
    DEFAULT_SEPARATORS = "|^~\\&"
    FIRST_SEP_INDEX = 3
    SEGMENT_ENDING = "\r"
    OPTIONS = [
        AstmMessageV2.Option(
            "cobas6800_prefix_qc", label="Préfixe QC dans le numéro de tube", default=""
        )
    ]

    def __init__(self, **kwargs):
        self.prefix_qc = kwargs.pop("cobas6800_prefix_qc", "QC")
        super(Driver, self).__init__(**kwargs)

    def create(self, **kwargs):
        return Cobas6800SegmentMSH(type_segment="MSH", **kwargs)

    def to_string(self, message, separators=None, **kwargs):
        return super(Driver, self).to_string(message, separators, **kwargs) + "\r"

    def encode_worklist(self, worklist, context):
        control_id = control_id = "MSG{}".format(
            context.get("control_id", int(time.time() * 1000))
        )
        message = self.create(date=context["now"], control_id=control_id)
        for patient in worklist.get_astm_worklist(
            unique_together=["action", "priority"],
            priority={"A": "S", "S": "S", worklist.ELSE: "R"},
            action={"del": "CA", worklist.ELSE: "NW"},
        ):
            for sample, action, priority in patient.samples:
                for order in sample["orders"]:
                    seg_spm = message.specimen.add(
                        sample_id=sample.sample_id, sample_identifier=sample.specimen
                    )
                    seg_sac = seg_spm.specimen_container.add()
                    seg_orc = seg_sac.common_order.add(order_control=action)
                    seg_obr = seg_orc.orders.add(test=order.code)
                    seg_tcd = seg_obr.infos_tests.add(test=order.code)
                    if order.dilution:
                        seg_tcd.pool_size_num1 = order.dilution
        return message

    def decode_message(self, message):
        if message.queries:
            worklist = Worklist(context="query")
            for seg_qpd in message.queries:
                sample = worklist.get_or_create(
                    type="patient", sample_id=seg_qpd.sample_identification
                )
        else:
            worklist = Worklist(context="result")
            specimens = message.specimen
            for seg_spm in specimens:
                if seg_spm.sample_role == "Q" or seg_spm.sample_id.startswith(
                    self.prefix_qc
                ):
                    sample = worklist.get_or_create(
                        sample_id=seg_spm.sample_id,
                        control_lot=seg_spm.sample_id,
                        type="control",
                    )
                else:
                    sample = worklist.get_or_create(
                        sample_id=seg_spm.sample_id, type="patient"
                    )
                for seg_sac in seg_spm.specimen_container:
                    sample.rack = seg_sac.container_identifier
                    for seg_obr in seg_sac.orders:
                        for seg_obx in seg_obr.results:
                            comments = []
                            for seg_nte in seg_obx.infos:
                                comments.append(seg_nte.comment)
                            result = sample.results.add(
                                code=seg_obx.observation_identifier,
                                value=seg_obx.value,
                                flags=[seg_obx.result_status],
                                operator=seg_obx.responsible_observer,
                                instrument=seg_obx.equipment_name,
                                unit=seg_obx.units_identifier,
                                date=seg_obx.analysis_date,
                                comments=comments,
                            )
                            for seg_inv in seg_obx.inventory:
                                result.lots.add(
                                    type="reagent", lot=seg_inv.fluid_identifier
                                )

        return worklist
